import logging
from datetime import datetime
from time import sleep

import event_chain.db.utils as db
from event_chain.application import did_auth
from event_chain.application import models
from event_chain.utils import helpers
from forge.rpc import rpc as forge_rpc

logger = logging.getLogger('ec-app')


def connect(db_path):
    conn = db.create_connection()
    logger.info("DB connected: {}.".format(db_path))
    return conn


def register_user(moniker, passphrase, conn=None):
    user = models.User(moniker, passphrase)
    logger.info("User {} created successfully!".format(moniker))
    user.poke()
    if conn:
        db.insert_user(conn, user.address, moniker, passphrase)
    return user


def load_user(moniker, passphrase, conn=None, address=None):
    address = address if address else db.select_address_by_moniker(
        conn,
        moniker,
    )
    user = models.User(moniker=moniker, passphrase=passphrase, address=address)
    logger.info("User {} loaded successfully!".format(moniker))
    return user


def parse_date(str_date):
    logger.debug(str_date)
    data = str_date.split('/')
    return datetime(
        int(data[0]),
        int(data[1]),
        int(data[2]),
    )


def create_event(user, conn=None, **kwargs):
    random_img = 'https://unsplash.it/800/450/'
    event_info = models.EventInfo(
        wallet=user.get_wallet(),
        token=user.token,
        title=kwargs.get('title'),
        total=int(kwargs.get('total')),
        description=kwargs.get('description'),
        start_time=parse_date(kwargs.get('start_time')),
        end_time=parse_date(kwargs.get('end_time')),
        ticket_price=int(kwargs.get('ticket_price')),
        location=kwargs.get('location'),
        img_url=kwargs.get('img_url', random_img)
    )
    if event_info.finished and conn:
        db.insert_event(conn, event_info.address, user.address)
        logger.debug(
            'Event {} has been added to database!'.format(
                event_info.address,
            ),
        )
    return event_info.address


def list_events(conn):
    addr_list = db.select_all_events(conn)
    event_states = []
    for addr in addr_list:
        state = get_event_state(addr)
        if state:
            event_states.append(state)
    return event_states


def list_event_exchange_txs(conn, event_address):
    hashes = db.select_exchange_txs(conn, event_address)
    tx_list = []
    for hash in hashes:
        tx = get_tx_info(hash)
        if tx:
            tx_list.append(tx)
    return tx_list


def list_event_detail(addr):
    event = get_event_state(addr)
    event_info = event.event_info
    print("****************************")
    print('Title:', event_info.title)
    print('Address', event.address)
    print('Total Tickets: ', event_info.total)
    print("Created by: ", event.owner)
    print("Remaining ticket number: ", event_info.remaining)


class TransactionInfo:
    def __init__(self, state):
        self.height = state.height
        self.hash = state.hash
        self.tx = state.tx
        self.time = state.time


def get_tx_info(hash):
    info = forge_rpc.get_single_tx_info(hash)
    if info:
        return TransactionInfo(info)


def list_unused_tickets(user_address):
    user_state = get_participant_state(user_address)
    if not user_state:
        return []
    else:
        addr_list = user_state.unused
        ticket_states = [
            get_ticket_state(addr) for addr in addr_list if
            get_ticket_state(addr) is not None
        ]
        return ticket_states


def get_event_state(event_address):
    return models.get_event_state(event_address)


def get_ticket_state(ticket_address):
    return models.get_ticket_state(ticket_address)


def get_participant_state(address):
    return models.get_participant_state(address)


def buy_ticket(event_address, user, conn=None):
    event_asset = get_event_state(event_address)
    logger.debug('user wallet: {}'.format(user.get_wallet()))
    logger.debug('user token: {}'.format(user.token))
    exchange_hash = event_asset.buy_ticket(
        user.get_wallet(), user.token,
    )
    logger.debug("Buy ticket process is completed. exchange hash{}".format(
        exchange_hash,
    ))
    if exchange_hash and conn:
        db.insert_exchange_tx(conn, event_address, exchange_hash)
    return exchange_hash


def buy_ticket_mobile(event_address, address, signature, user_pk, conn=None):
    state = get_event_state(event_address)
    ticket_address, hash = state.buy_ticket_mobile(
        address, signature, user_pk
    )
    if hash and conn:
        db.insert_exchange_tx(conn, event_address, hash)
    return ticket_address, hash


def get_wallet_address(response):
    wallet_response = helpers.WalletResponse(response)
    address = wallet_response.get_address()
    return address


def create_sample_event(user, title, conn=None):
    return create_event(
        user=user,
        conn=conn,
        title=title,
        total=20,
        description='This is a sample event created for demo',
        start_time=datetime(2019, 2, 9, 21),
        end_time=datetime(2019, 2, 10, 23),
        ticket_price=188,
    )


def consume(ticket_address, user):
    logger.debug("Consuming ticket {}".format(ticket_address))
    ticket = get_ticket_state(ticket_address)
    logger.debug("Event is  {}".format(ticket.ticket_info.event_address))
    consume_tx = get_event_state(
        ticket.ticket_info.event_address,
    ).event_info.consume_tx
    if not consume_tx:
        return None

    logger.debug("consume tx: {}".format(consume_tx))

    res = ticket.consume(consume_tx, user.get_wallet(), user.token)

    if res.code != 0 or res.hash is None:
        logger.error(res)
        logger.error('Fail to consume ticket {}'.format(ticket_address))
    else:
        logger.info(
            "ConsumeTx has been sent by tx: {}!".format(res.hash),
        )
    return res.hash


def consume_ticket_mobile(ticket, consume_tx, address, signature, user_pk):
    res = ticket.consume_mobile(consume_tx, address, signature, user_pk)

    if res.code != 0 or res.hash is None:
        logger.error(res)
        logger.error(
            'Fail to consume ticket by mobile {}'.format(ticket.address),
        )
    else:
        logger.info("Mobile ConsumeTx has been sent by tx: {}!".format(
            res.hash,
        ))
    return res.hash


def list_ticket_exchange_tx(conn, event_address):
    tx_list = []
    res = db.select_exchange_txs(conn, event_address)
    for tx in res.transactions:
        if tx.type == 'exchange':
            tx_list.append(tx)
    return tx_list


def verify_event_address(event_address):
    try:
        state = get_event_state(event_address)
    except Exception:
        logger.error('exception in verifying event_address ')
        raise TypeError("{} is not an event address.".format(event_address))
    if not state:
        logger.error('Event {} does not exist.'.format(event_address))
        raise ValueError('Event {} does not exist'.format(event_address))
    if state.type_url != 'ec:s:event_info':
        logger.error(
            'This asset should be type ec:s:event_info, but provided '
            'type_url is : {}'.format(
                state.type_url,
            ),
        )
        raise ValueError(
            '{} is not a valid event address. Type_url for this asset '
            'is {}'.format(
                event_address, state.type_url,
            ),
        )


def verify_ticket_address(ticket_address):
    try:
        state = get_ticket_state(ticket_address)
    except Exception:
        logger.error('Error in checking ticket')
        raise TypeError("{} is not an ticket address.".format(ticket_address))

    if not state:
        logger.error(u'Ticket {} does not exist.'.format(ticket_address))
        raise ValueError('Ticket {} does not exist'.format(ticket_address))
    if state.type_url != 'ec:s:ticket_info':
        logger.error('This asset type should be ec:s:ticket_info,'
                     ' but the provided url is: {}'.format(state.type_url))
        raise ValueError(
            '{} is not a valid ticket address. Type_url for this asset '
            'is {}'.format(
                ticket_address, state.type_url,
            ),
        )


def did_auth_require_multisig(**kwargs):
    return did_auth.response_require_multisig(**kwargs)


def did_auth_require_asset(**kwargs):
    return did_auth.response_require_asset(**kwargs)


def refresh():
    sleep(5)



import hashlib

import sha3


def is_sha2(name):
    return name.startswith('sha2')


def parse_hash_type(name):
    if name == 'sha2':
        return 'sha256'
    elif name.startswith('sha2_'):
        return name.replace('sha2_', 'sha')
    elif name == 'sha3':
        return 'sha3_256'
    elif name == 'keccak':
        return 'keccak_256'
    else:
        return name


class Hasher:
    sizes = [224, 256, 384, 512]
    types = ['sha2', 'sha3', 'keccak']

    def __init__(self, name, round=None):
        self.name = name
        self.supported_hash = [t + '_' + str(z) for t in Hasher.types
                               for z in Hasher.sizes] + Hasher.types

        if is_sha2(name) and not round:
            self.round = 2
        else:
            self.round = round if round else 1

    def is_supported(self, name):
        if name in self.supported_hash:
            return True
        else:
            return False

    def init_hasher(self, name):
        if not self.is_supported(name):
            raise NameError("{} hash type is not supported.".format(name))

        hash_type = parse_hash_type(name)
        if '_' not in hash_type:
            res = getattr(hashlib, hash_type)
        else:
            res = getattr(sha3, hash_type)
        return res()

    def hash(self, data_bytes):
        res = data_bytes
        for i in range(0, self.round):
            hash_helper = self.init_hasher(self.name)
            hash_helper.update(res)
            res = hash_helper.digest()
        return res



from forge import protos
from forge.config import config

stub = protos.StatisticRpcStub(config.get_grpc_channel())


def list_asset_transactions(address, paging=None):
    return stub.list_asset_transactions(
        protos.RequestListAssetTransactions(
            address=address,
            paging=paging,
        ),
    )


def list_transactions(
    address_filter, time_filter=None,
        type_filter=None, paging=None,
):
    req = protos.RequestListTransactions(
        paging=paging,
        time_filter=time_filter,
        address_filter=address_filter,
        type_filter=type_filter,
    )
    return stub.list_transactions(req)



import logging

from .config import ForgeConfig
from .server import ForgeServer


class ForgeSdk:
    def __init__(self, handlers=[], config=None):
        logging.basicConfig(
            format='%(asctime)s-%(name)s-%(process)d-%('
            'levelname)s-%(message)s',
        )
        self.handlers = handlers
        self.logger = logging.getLogger(__name__)
        self.config = config if config else ForgeConfig()
        self.rpc = ForgeRpc(self.config.sock_grpc)

        self.server = ForgeServer(
            handlers=handlers,
            address=self.config.sock_tcp,
        )

    def register_handler(self, handler):
        # sanity check
        # add to server
        self.server.register_handler(handler)

    def start(self):
        self.server.start()


class TxHandler:
    def __init__(self, tx_type, function_dic):
        self.tx_type = tx_type
        self.function_dic = function_dic



from forge import protos
from forge.utils import utils


class RpcChain:
    def __init__(self, chan):
        self.stub = protos.ChainRpcStub(chan)

    def create_tx(
            self, itx=None, from_address='',
            wallet=None, token=None, req=None, nonce=1,
    ):
        """
        RPC call to create transaction.

        Parameters
        ----------
        req : RequestCreateTx
        itx : google.protobuf.An
        from_address : string
        nonce : uint64
        wallet : WalletInfo
        token : string

        Returns
        -------
        ResponseCreateTx

        """

        if req is not None:
            return self.stub.create_tx(req)
        else:
            req_kwargs = {
                'itx': itx,
                'from': from_address,
                'nonce': nonce,
                'wallet': wallet,
                'token': token,
            }
            return self.stub.create_tx(protos.RequestCreateTx(**req_kwargs))

    def send_tx(
            self, tx=None, wallet=None, token=None, commit=False, req=None,
    ):
        """
        RPC call to send transaction.

        Parameters
        ----------
        req: RequestSendTx
        tx : Transaction
        wallet: WalletInfo
        token: string
        commit: bool

        Returns
        -------
        ResponseSendTx

        """
        if req is not None:
            return self.stub.send_tx(req)
        else:
            req_kwargs = {
                'tx': tx,
                'wallet': wallet,
                'token': token,
                'commit': commit,
            }
            req = protos.RequestSendTx(**req_kwargs)
            return self.stub.send_tx(req)

    def get_tx(self, tx_hash='', req=None):
        """
        RPC call to get transaction.

        Parameters
        ----------
        req: stream RequestGetTx
        tx_hash: single string or string iterator

        Returns
        -------
        stream  ResponseGetTx

        """

        def to_req(item):
            if isinstance(item, protos.RequestGetTx):
                return item
            else:
                return protos.RequestGetTx(hash=item)

        if req is not None:
            return self.stub.get_tx(utils.to_iter(to_req, req))
        else:
            return self.stub.get_tx(utils.to_iter(to_req, tx_hash))

    def get_block(self, height=0, req=None):
        """
        RPC call to get blocks.

        Parameters
        ----------
        req: RequestGetBlock
        height: uint64

        Returns
        -------
        stream ResponseGetBlock

        """

        def to_req(item):
            if isinstance(item, protos.RequestGetBlock):
                return item
            else:
                return protos.RequestGetBlock(height=item)

        if req is not None:
            return self.stub.get_block(utils.to_iter(to_req, req))
        else:
            return self.stub.get_block(utils.to_iter(to_req, height))

    def search(self, key='', value='', req=None):
        """

        Parameters
        ----------
        req: RequestSearch
        key: string
        value: string

        Returns
        -------
        ResponseSearch

        """
        if req is not None:
            return self.stub.search(req)
        else:
            req_kwargs = {
                'key': key,
                'value': value,
            }
            return self.stub.search(protos.RequestSearch(**req_kwargs))

    def get_unconfirmed_tx(self, req=None, limit=1):
        """

        Parameters
        ----------
        req: RequestGetUnconfirmedTxs
        limit: int

        Returns
        -------
        ResponseGetUnconfirmedTxs

        """
        if req is not None:
            return self.stub.get_unconfirmed_txs(req)
        else:
            return self.stub.get_unconfirmed_txs(
                protos.RequestGetUnconfirmedTxs(limit=limit),
            )

    def get_chain_info(self, req=None):
        """

        Parameters
        ----------
        req: RequestGetChainInfo

        Returns
        -------
        ResponseGetChainInfo

        """
        if req is not None:
            return self.stub.get_chain_info(req)
        else:
            return self.stub.get_chain_info(protos.RequestGetChainInfo())

    def get_net_info(self, req=None):
        """

        Parameters
        ----------
        req: RequestGetNetInfo

        Returns
        -------
        ResponseGetNetInfo

        """
        if req is not None:
            return self.stub.get_net_info(req)
        else:
            return self.stub.get_net_info(protos.RequestGetNetInfo())

    def get_validators_info(self, req=None):
        """

        Parameters
        ----------
        req: RequestGetValidatorsInfo

        Returns
        -------
        ResponseGetValidatorsInfo

        """
        if req is not None:
            return self.stub.get_validators_info(req)
        else:
            return self.stub.get_validators_info(
                protos.RequestGetValidatorsInfo(),
            )

    def get_config(self, req=None):
        """

        Parameters
        ----------
        req: RequestGetConfig

        Returns
        -------
        ResponseGetConfig

        """
        if req is not None:
            return self.stub.get_config(req)
        else:
            return self.stub.get_config(
                protos.RequestGetConfig(),
            )

    def multisig(self, tx=None, wallet=None, token=None, data=None, req=None):

        if req is not None:
            return self.stub.multisig(req)
        else:
            req = protos.RequestMultisig(
                tx=tx, wallet=wallet, token=token,
                data=data,
            )
            return self.stub.multisig(req)

    def get_asset_address(
            self, sender_address='', itx=None,
            wallet_type=None, req=None,
    ):
        if req is not None:
            return self.stub.get_asset_address(req)
        else:
            return self.stub.get_asset_address(
                protos.RequestGetAssetAddress(
                    sender_address=sender_address, itx=itx,
                    wallet_type=wallet_type,
                ),
            )



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tx.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from . import type_pb2 as type__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tx.proto',
  package='forge_abi',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x08tx.proto\x12\tforge_abi\x1a\x19google/protobuf/any.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\ntype.proto\"g\n\x10\x41\x63\x63ountMigrateTx\x12\n\n\x02pk\x18\x01 \x01(\x0c\x12#\n\x04type\x18\x02 \x01(\x0b\x32\x15.forge_abi.WalletType\x12\"\n\x04\x64\x61ta\x18\x0f \x01(\x0b\x32\x14.google.protobuf.Any\"\xb6\x01\n\x12\x43onsensusUpgradeTx\x12(\n\nvalidators\x18\x01 \x03(\x0b\x32\x14.forge_abi.Validator\x12\x11\n\tmax_bytes\x18\x02 \x01(\x04\x12\x0f\n\x07max_gas\x18\x03 \x01(\x12\x12\x16\n\x0emax_validators\x18\x04 \x01(\r\x12\x16\n\x0emax_candidates\x18\x05 \x01(\r\x12\"\n\x04\x64\x61ta\x18\x0f \x01(\x0b\x32\x14.google.protobuf.Any\"U\n\x0e\x43onsumeAssetTx\x12\x0e\n\x06issuer\x18\x01 \x01(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\t\x12\"\n\x04\x64\x61ta\x18\x0f \x01(\x0b\x32\x14.google.protobuf.Any\"\x8a\x01\n\rCreateAssetTx\x12\x0f\n\x07moniker\x18\x01 \x01(\t\x12\"\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any\x12\x10\n\x08readonly\x18\x03 \x01(\x08\x12\x15\n\rtransferrable\x18\x04 \x01(\x08\x12\x0b\n\x03ttl\x18\x05 \x01(\r\x12\x0e\n\x06parent\x18\x06 \x01(\t\"P\n\tDeclareTx\x12\x0f\n\x07moniker\x18\x01 \x01(\t\x12\x0e\n\x06issuer\x18\x02 \x01(\t\x12\"\n\x04\x64\x61ta\x18\x0f \x01(\x0b\x32\x14.google.protobuf.Any\"\x1d\n\rDeclareFileTx\x12\x0c\n\x04hash\x18\x01 \x01(\t\"A\n\x0c\x45xchangeInfo\x12!\n\x05value\x18\x01 \x01(\x0b\x32\x12.forge_abi.BigUint\x12\x0e\n\x06\x61ssets\x18\x02 \x03(\t\"\xc0\x01\n\nExchangeTx\x12\n\n\x02to\x18\x01 \x01(\t\x12\'\n\x06sender\x18\x02 \x01(\x0b\x32\x17.forge_abi.ExchangeInfo\x12)\n\x08receiver\x18\x03 \x01(\x0b\x32\x17.forge_abi.ExchangeInfo\x12.\n\nexpired_at\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\"\n\x04\x64\x61ta\x18\x0f \x01(\x0b\x32\x14.google.protobuf.Any\"\x0f\n\rstakeForAsset\"\x0f\n\rstakeForChain\"\x0e\n\x0cStakeForNode\"\x0e\n\x0cstakeForUser\"m\n\x07StakeTx\x12\n\n\x02to\x18\x01 \x01(\t\x12!\n\x05value\x18\x02 \x01(\x0b\x32\x12.forge_abi.BigSint\x12\x0f\n\x07message\x18\x03 \x01(\t\x12\"\n\x04\x64\x61ta\x18\x0f \x01(\x0b\x32\x14.google.protobuf.Any\"n\n\x0cSysUpgradeTx\x12$\n\x04task\x18\x01 \x01(\x0b\x32\x16.forge_abi.UpgradeTask\x12\x14\n\x0cgrace_period\x18\x02 \x01(\x04\x12\"\n\x04\x64\x61ta\x18\x0f \x01(\x0b\x32\x14.google.protobuf.Any\"o\n\nTransferTx\x12\n\n\x02to\x18\x01 \x01(\t\x12!\n\x05value\x18\x02 \x01(\x0b\x32\x12.forge_abi.BigUint\x12\x0e\n\x06\x61ssets\x18\x03 \x03(\t\x12\"\n\x04\x64\x61ta\x18\x0f \x01(\x0b\x32\x14.google.protobuf.Any\"U\n\rUpdateAssetTx\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\x12\x0f\n\x07moniker\x18\x02 \x01(\t\x12\"\n\x04\x64\x61ta\x18\x0f \x01(\x0b\x32\x14.google.protobuf.Any\"K\n\x06PokeTx\x12\x0c\n\x04\x64\x61te\x18\x01 \x01(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\t\x12\"\n\x04\x64\x61ta\x18\x0f \x01(\x0b\x32\x14.google.protobuf.Anyb\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_any__pb2.DESCRIPTOR,google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,type__pb2.DESCRIPTOR,])




_ACCOUNTMIGRATETX = _descriptor.Descriptor(
  name='AccountMigrateTx',
  full_name='forge_abi.AccountMigrateTx',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pk', full_name='forge_abi.AccountMigrateTx.pk', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='forge_abi.AccountMigrateTx.type', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='forge_abi.AccountMigrateTx.data', index=2,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=95,
  serialized_end=198,
)


_CONSENSUSUPGRADETX = _descriptor.Descriptor(
  name='ConsensusUpgradeTx',
  full_name='forge_abi.ConsensusUpgradeTx',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='validators', full_name='forge_abi.ConsensusUpgradeTx.validators', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_bytes', full_name='forge_abi.ConsensusUpgradeTx.max_bytes', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_gas', full_name='forge_abi.ConsensusUpgradeTx.max_gas', index=2,
      number=3, type=18, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_validators', full_name='forge_abi.ConsensusUpgradeTx.max_validators', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_candidates', full_name='forge_abi.ConsensusUpgradeTx.max_candidates', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='forge_abi.ConsensusUpgradeTx.data', index=5,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=201,
  serialized_end=383,
)


_CONSUMEASSETTX = _descriptor.Descriptor(
  name='ConsumeAssetTx',
  full_name='forge_abi.ConsumeAssetTx',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='issuer', full_name='forge_abi.ConsumeAssetTx.issuer', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='address', full_name='forge_abi.ConsumeAssetTx.address', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='forge_abi.ConsumeAssetTx.data', index=2,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=385,
  serialized_end=470,
)


_CREATEASSETTX = _descriptor.Descriptor(
  name='CreateAssetTx',
  full_name='forge_abi.CreateAssetTx',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='moniker', full_name='forge_abi.CreateAssetTx.moniker', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='forge_abi.CreateAssetTx.data', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='readonly', full_name='forge_abi.CreateAssetTx.readonly', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transferrable', full_name='forge_abi.CreateAssetTx.transferrable', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ttl', full_name='forge_abi.CreateAssetTx.ttl', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='parent', full_name='forge_abi.CreateAssetTx.parent', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=473,
  serialized_end=611,
)


_DECLARETX = _descriptor.Descriptor(
  name='DeclareTx',
  full_name='forge_abi.DeclareTx',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='moniker', full_name='forge_abi.DeclareTx.moniker', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='issuer', full_name='forge_abi.DeclareTx.issuer', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='forge_abi.DeclareTx.data', index=2,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=613,
  serialized_end=693,
)


_DECLAREFILETX = _descriptor.Descriptor(
  name='DeclareFileTx',
  full_name='forge_abi.DeclareFileTx',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='hash', full_name='forge_abi.DeclareFileTx.hash', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=695,
  serialized_end=724,
)


_EXCHANGEINFO = _descriptor.Descriptor(
  name='ExchangeInfo',
  full_name='forge_abi.ExchangeInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='forge_abi.ExchangeInfo.value', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='assets', full_name='forge_abi.ExchangeInfo.assets', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=726,
  serialized_end=791,
)


_EXCHANGETX = _descriptor.Descriptor(
  name='ExchangeTx',
  full_name='forge_abi.ExchangeTx',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='to', full_name='forge_abi.ExchangeTx.to', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sender', full_name='forge_abi.ExchangeTx.sender', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='receiver', full_name='forge_abi.ExchangeTx.receiver', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='expired_at', full_name='forge_abi.ExchangeTx.expired_at', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='forge_abi.ExchangeTx.data', index=4,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=794,
  serialized_end=986,
)


_STAKEFORASSET = _descriptor.Descriptor(
  name='stakeForAsset',
  full_name='forge_abi.stakeForAsset',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=988,
  serialized_end=1003,
)


_STAKEFORCHAIN = _descriptor.Descriptor(
  name='stakeForChain',
  full_name='forge_abi.stakeForChain',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1005,
  serialized_end=1020,
)


_STAKEFORNODE = _descriptor.Descriptor(
  name='StakeForNode',
  full_name='forge_abi.StakeForNode',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1022,
  serialized_end=1036,
)


_STAKEFORUSER = _descriptor.Descriptor(
  name='stakeForUser',
  full_name='forge_abi.stakeForUser',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1038,
  serialized_end=1052,
)


_STAKETX = _descriptor.Descriptor(
  name='StakeTx',
  full_name='forge_abi.StakeTx',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='to', full_name='forge_abi.StakeTx.to', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='forge_abi.StakeTx.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message', full_name='forge_abi.StakeTx.message', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='forge_abi.StakeTx.data', index=3,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1054,
  serialized_end=1163,
)


_SYSUPGRADETX = _descriptor.Descriptor(
  name='SysUpgradeTx',
  full_name='forge_abi.SysUpgradeTx',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='task', full_name='forge_abi.SysUpgradeTx.task', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='grace_period', full_name='forge_abi.SysUpgradeTx.grace_period', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='forge_abi.SysUpgradeTx.data', index=2,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1165,
  serialized_end=1275,
)


_TRANSFERTX = _descriptor.Descriptor(
  name='TransferTx',
  full_name='forge_abi.TransferTx',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='to', full_name='forge_abi.TransferTx.to', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='forge_abi.TransferTx.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='assets', full_name='forge_abi.TransferTx.assets', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='forge_abi.TransferTx.data', index=3,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1277,
  serialized_end=1388,
)


_UPDATEASSETTX = _descriptor.Descriptor(
  name='UpdateAssetTx',
  full_name='forge_abi.UpdateAssetTx',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='forge_abi.UpdateAssetTx.address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='moniker', full_name='forge_abi.UpdateAssetTx.moniker', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='forge_abi.UpdateAssetTx.data', index=2,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1390,
  serialized_end=1475,
)


_POKETX = _descriptor.Descriptor(
  name='PokeTx',
  full_name='forge_abi.PokeTx',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='date', full_name='forge_abi.PokeTx.date', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='address', full_name='forge_abi.PokeTx.address', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='forge_abi.PokeTx.data', index=2,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1477,
  serialized_end=1552,
)

_ACCOUNTMIGRATETX.fields_by_name['type'].message_type = type__pb2._WALLETTYPE
_ACCOUNTMIGRATETX.fields_by_name['data'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_CONSENSUSUPGRADETX.fields_by_name['validators'].message_type = type__pb2._VALIDATOR
_CONSENSUSUPGRADETX.fields_by_name['data'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_CONSUMEASSETTX.fields_by_name['data'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_CREATEASSETTX.fields_by_name['data'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_DECLARETX.fields_by_name['data'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_EXCHANGEINFO.fields_by_name['value'].message_type = type__pb2._BIGUINT
_EXCHANGETX.fields_by_name['sender'].message_type = _EXCHANGEINFO
_EXCHANGETX.fields_by_name['receiver'].message_type = _EXCHANGEINFO
_EXCHANGETX.fields_by_name['expired_at'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_EXCHANGETX.fields_by_name['data'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_STAKETX.fields_by_name['value'].message_type = type__pb2._BIGSINT
_STAKETX.fields_by_name['data'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_SYSUPGRADETX.fields_by_name['task'].message_type = type__pb2._UPGRADETASK
_SYSUPGRADETX.fields_by_name['data'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_TRANSFERTX.fields_by_name['value'].message_type = type__pb2._BIGUINT
_TRANSFERTX.fields_by_name['data'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_UPDATEASSETTX.fields_by_name['data'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_POKETX.fields_by_name['data'].message_type = google_dot_protobuf_dot_any__pb2._ANY
DESCRIPTOR.message_types_by_name['AccountMigrateTx'] = _ACCOUNTMIGRATETX
DESCRIPTOR.message_types_by_name['ConsensusUpgradeTx'] = _CONSENSUSUPGRADETX
DESCRIPTOR.message_types_by_name['ConsumeAssetTx'] = _CONSUMEASSETTX
DESCRIPTOR.message_types_by_name['CreateAssetTx'] = _CREATEASSETTX
DESCRIPTOR.message_types_by_name['DeclareTx'] = _DECLARETX
DESCRIPTOR.message_types_by_name['DeclareFileTx'] = _DECLAREFILETX
DESCRIPTOR.message_types_by_name['ExchangeInfo'] = _EXCHANGEINFO
DESCRIPTOR.message_types_by_name['ExchangeTx'] = _EXCHANGETX
DESCRIPTOR.message_types_by_name['stakeForAsset'] = _STAKEFORASSET
DESCRIPTOR.message_types_by_name['stakeForChain'] = _STAKEFORCHAIN
DESCRIPTOR.message_types_by_name['StakeForNode'] = _STAKEFORNODE
DESCRIPTOR.message_types_by_name['stakeForUser'] = _STAKEFORUSER
DESCRIPTOR.message_types_by_name['StakeTx'] = _STAKETX
DESCRIPTOR.message_types_by_name['SysUpgradeTx'] = _SYSUPGRADETX
DESCRIPTOR.message_types_by_name['TransferTx'] = _TRANSFERTX
DESCRIPTOR.message_types_by_name['UpdateAssetTx'] = _UPDATEASSETTX
DESCRIPTOR.message_types_by_name['PokeTx'] = _POKETX
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

AccountMigrateTx = _reflection.GeneratedProtocolMessageType('AccountMigrateTx', (_message.Message,), dict(
  DESCRIPTOR = _ACCOUNTMIGRATETX,
  __module__ = 'tx_pb2'
  # @@protoc_insertion_point(class_scope:forge_abi.AccountMigrateTx)
  ))
_sym_db.RegisterMessage(AccountMigrateTx)

ConsensusUpgradeTx = _reflection.GeneratedProtocolMessageType('ConsensusUpgradeTx', (_message.Message,), dict(
  DESCRIPTOR = _CONSENSUSUPGRADETX,
  __module__ = 'tx_pb2'
  # @@protoc_insertion_point(class_scope:forge_abi.ConsensusUpgradeTx)
  ))
_sym_db.RegisterMessage(ConsensusUpgradeTx)

ConsumeAssetTx = _reflection.GeneratedProtocolMessageType('ConsumeAssetTx', (_message.Message,), dict(
  DESCRIPTOR = _CONSUMEASSETTX,
  __module__ = 'tx_pb2'
  # @@protoc_insertion_point(class_scope:forge_abi.ConsumeAssetTx)
  ))
_sym_db.RegisterMessage(ConsumeAssetTx)

CreateAssetTx = _reflection.GeneratedProtocolMessageType('CreateAssetTx', (_message.Message,), dict(
  DESCRIPTOR = _CREATEASSETTX,
  __module__ = 'tx_pb2'
  # @@protoc_insertion_point(class_scope:forge_abi.CreateAssetTx)
  ))
_sym_db.RegisterMessage(CreateAssetTx)

DeclareTx = _reflection.GeneratedProtocolMessageType('DeclareTx', (_message.Message,), dict(
  DESCRIPTOR = _DECLARETX,
  __module__ = 'tx_pb2'
  # @@protoc_insertion_point(class_scope:forge_abi.DeclareTx)
  ))
_sym_db.RegisterMessage(DeclareTx)

DeclareFileTx = _reflection.GeneratedProtocolMessageType('DeclareFileTx', (_message.Message,), dict(
  DESCRIPTOR = _DECLAREFILETX,
  __module__ = 'tx_pb2'
  # @@protoc_insertion_point(class_scope:forge_abi.DeclareFileTx)
  ))
_sym_db.RegisterMessage(DeclareFileTx)

ExchangeInfo = _reflection.GeneratedProtocolMessageType('ExchangeInfo', (_message.Message,), dict(
  DESCRIPTOR = _EXCHANGEINFO,
  __module__ = 'tx_pb2'
  # @@protoc_insertion_point(class_scope:forge_abi.ExchangeInfo)
  ))
_sym_db.RegisterMessage(ExchangeInfo)

ExchangeTx = _reflection.GeneratedProtocolMessageType('ExchangeTx', (_message.Message,), dict(
  DESCRIPTOR = _EXCHANGETX,
  __module__ = 'tx_pb2'
  # @@protoc_insertion_point(class_scope:forge_abi.ExchangeTx)
  ))
_sym_db.RegisterMessage(ExchangeTx)

stakeForAsset = _reflection.GeneratedProtocolMessageType('stakeForAsset', (_message.Message,), dict(
  DESCRIPTOR = _STAKEFORASSET,
  __module__ = 'tx_pb2'
  # @@protoc_insertion_point(class_scope:forge_abi.stakeForAsset)
  ))
_sym_db.RegisterMessage(stakeForAsset)

stakeForChain = _reflection.GeneratedProtocolMessageType('stakeForChain', (_message.Message,), dict(
  DESCRIPTOR = _STAKEFORCHAIN,
  __module__ = 'tx_pb2'
  # @@protoc_insertion_point(class_scope:forge_abi.stakeForChain)
  ))
_sym_db.RegisterMessage(stakeForChain)

StakeForNode = _reflection.GeneratedProtocolMessageType('StakeForNode', (_message.Message,), dict(
  DESCRIPTOR = _STAKEFORNODE,
  __module__ = 'tx_pb2'
  # @@protoc_insertion_point(class_scope:forge_abi.StakeForNode)
  ))
_sym_db.RegisterMessage(StakeForNode)

stakeForUser = _reflection.GeneratedProtocolMessageType('stakeForUser', (_message.Message,), dict(
  DESCRIPTOR = _STAKEFORUSER,
  __module__ = 'tx_pb2'
  # @@protoc_insertion_point(class_scope:forge_abi.stakeForUser)
  ))
_sym_db.RegisterMessage(stakeForUser)

StakeTx = _reflection.GeneratedProtocolMessageType('StakeTx', (_message.Message,), dict(
  DESCRIPTOR = _STAKETX,
  __module__ = 'tx_pb2'
  # @@protoc_insertion_point(class_scope:forge_abi.StakeTx)
  ))
_sym_db.RegisterMessage(StakeTx)

SysUpgradeTx = _reflection.GeneratedProtocolMessageType('SysUpgradeTx', (_message.Message,), dict(
  DESCRIPTOR = _SYSUPGRADETX,
  __module__ = 'tx_pb2'
  # @@protoc_insertion_point(class_scope:forge_abi.SysUpgradeTx)
  ))
_sym_db.RegisterMessage(SysUpgradeTx)

TransferTx = _reflection.GeneratedProtocolMessageType('TransferTx', (_message.Message,), dict(
  DESCRIPTOR = _TRANSFERTX,
  __module__ = 'tx_pb2'
  # @@protoc_insertion_point(class_scope:forge_abi.TransferTx)
  ))
_sym_db.RegisterMessage(TransferTx)

UpdateAssetTx = _reflection.GeneratedProtocolMessageType('UpdateAssetTx', (_message.Message,), dict(
  DESCRIPTOR = _UPDATEASSETTX,
  __module__ = 'tx_pb2'
  # @@protoc_insertion_point(class_scope:forge_abi.UpdateAssetTx)
  ))
_sym_db.RegisterMessage(UpdateAssetTx)

PokeTx = _reflection.GeneratedProtocolMessageType('PokeTx', (_message.Message,), dict(
  DESCRIPTOR = _POKETX,
  __module__ = 'tx_pb2'
  # @@protoc_insertion_point(class_scope:forge_abi.PokeTx)
  ))
_sym_db.RegisterMessage(PokeTx)


# @@protoc_insertion_point(module_scope)



from forge import protos
from forge.utils import utils


class RpcFile:
    def __init__(self, chan):
        self.stub = protos.FileRpcStub(chan)

    def store_file(self, chunk=b'', req=None):
        """
        RPC call to store file

        Parameters
        ----------
        req: stream RequestStoreFile
        chunk: iterator of bytes

        Returns
        -------
        ResponseStoreFile

        """

        def to_req(item):
            if isinstance(item, protos.RequestStoreFile):
                return item
            else:
                return protos.RequestStoreFile(chunk=item)

        if req is not None:
            return self.stub.store_file(utils.to_iter(to_req, req))
        else:
            chunks = utils.to_iter(to_req, chunk)
            return self.stub.store_file(chunks)

    def load_file(self, file_hash='', req=None):
        """
        RPC call to load file.

        Parameters
        ----------
        req: RequestLoadFile
        file_hash: string

        Returns
        -------
        stream ResponseLoadFile

        """
        if req is not None:
            return self.stub.load_file(req)
        else:
            req_kwargs = {
                'hash': file_hash,
            }
            return self.stub.load_file(protos.RequestLoadFile(**req_kwargs))



import logging

from forge.config import config
from forge.server import ForgeServer

logger = logging.getLogger('forge-sdk')

server = None


def init_server(handlers):
    global server
    if not server:
        server = ForgeServer(handlers, config.get_tcp_socket())



import logging
import os
import os.path as path
from os.path import expanduser

import toml
from deepmerge import Merger

logger = logging.getLogger('Forge-Parser')


def parse_config(file_path):
    default_config = path.join(path.dirname(__file__), "forge_default.toml")
    toml_dict = toml.load(default_config)
    custom_path = get_config_path(file_path)

    if not custom_path:
        return toml_dict
    else:
        merger = Merger(
            # pass in a list of tuple, with the
            # strategies you are looking to apply
            # to each type.
            [
                (list, ["append"]),
                (dict, ["merge"]),
            ],
            # next, choose the fallback strategies,
            # applied to all other types:
            ["override"],
            # finally, choose the strategies in
            # the case where the types conflict:
            ["override"],
        )

        user_dict = toml.load(custom_path)
        merger.merge(toml_dict, user_dict)
    return toml_dict


def get_config_path(user_file=None):
    env = os.environ.get('FORGE_CONFIG')
    if env:
        if path.exists(env):
            logger.info('Using config in {}'.format(env))
            return env
        else:
            logger.error('Config file {} not found.'.format(env))
    if user_file:
        if path.exists(user_file):
            logger.info('Using config in {}'.format(user_file))
            return user_file
        else:
            logger.error('Config file {} not found.'.format(user_file))


class ForgeConfig:

    def __init__(self, file_path=None):
        self.toml_dict = parse_config(file_path)
        self.forge_path = expanduser(self.toml_dict['forge']['path'])
        self.sock_grpc = self.__parse_socket_grpc(
            self.forge_path,
            self.toml_dict['forge']['sock_grpc'],
        )
        self.sock_tcp = self.__parse_socket(
            self.get_app_path(),
            self.toml_dict['app']['sock_tcp'],
        )

    def get_app_path(self):
        return expanduser(self.toml_dict['app']['path'])

    def get_app_host(self):
        return self.toml_dict['app']['host']

    def get_app_port(self):
        return self.toml_dict['app']['port']

    def __parse_socket(self, forge_path, forge_socket):
        """

        Parameters
        ----------
        Returns: string
            unix: "unix:///tmp/.forge_test/app/socks/abi.sock"
            tcp: "tcp://127.0.0.1:38210"

        -------

        """
        socket_type = forge_socket.split("://")[0]
        parsed_socket = forge_socket.split("://")[1]
        if socket_type == 'unix':
            socket_target = '/'.join([
                'unix:/',
                forge_path, parsed_socket,
            ])
        else:
            socket_target = forge_socket
        return socket_target

    def __parse_socket_grpc(self, forge_path, forge_socket):
        socket_addr = self.__parse_socket(forge_path, forge_socket)
        socket_type = socket_addr.split('://')[0]
        if socket_type == 'unix':
            return socket_addr
        elif socket_type == 'tcp':
            return socket_addr.split('://')[1]



import logging
import os
import os.path as path

from forge.config import config as forge_config

logger = logging.getLogger('ec-config')

app_config_path = path.join(path.dirname(__file__), "forge.toml")

forge_config.use_config(app_config_path)

app_path = forge_config.get_app_path()

app_host = forge_config.get_app_host()

db_path = path.join(app_path, "ec.db")

googlemaps_key = os.environ.get('GOOGLEMAPS_KEY')



from forge import protos
from forge import rpc
from forge.config import ForgeConfig
from forge.did import AbtDid
from forge.forge_sdk import ForgeSdk
from forge.mcrypto import Hasher
from forge.mcrypto import Signer
from forge.sdk import sdk as ForgeSDK
from forge.server import ForgeServer
from forge.utils import helper
from forge.utils import utils



import logging

from event_chain import protos
from event_chain.application import models

from forge import helper as forge_helper
from forge import utils as forge_utils
from forge.sdk import sdk

logger = logging.getLogger('ec-server')

INVALID_SENDER_STATE = protos.StatusCode.Value('invalid_sender_state')
OK = protos.StatusCode.Value('ok')


def exchange_verify(request):
    logger.debug("ExchangeTx verify_request has been received.")
    # TODO: should get type_url from parameters
    exchange_tx = forge_utils.parse_to_proto(
        request.tx.itx.value,
        protos.ExchangeTx,
    )
    if exchange_tx.data.type_url == "ec:x:event_address":
        event_address = exchange_tx.data.value.decode('utf8')
        logger.debug(
            "ExchangeTx contains an event address {}".format(
                event_address,
            ),
        )
    else:
        logger.error(
            "Received an verify_reqeust for ExchangeTx not recognized.",
        )

    return protos.ResponseVerifyTx(code=0)


def exchange_update(request):
    logger.debug("ExchangeTx update_request has been received.")
    exchange_tx = forge_utils.parse_to_proto(
        request.tx.itx.value,
        protos.ExchangeTx,
    )

    if exchange_tx.data.type_url == "ec:x:event_address":
        event_address = exchange_tx.data.value.decode()
        logger.debug(
            "ExchangeTx contains an event address {}".format(
                event_address,
            ),
        )
        # update event state
        asset_state = models.get_event_state(event_address)
        updated_event_state = asset_state.pop_executed_ticket()
        updated_event_info = forge_utils.parse_to_proto(
            updated_event_state.data.value,
            protos.EventInfo,
        )

        # update buyer state
        buyer_address = request.tx.signatures[0].signer
        logger.debug("buyer address :{}".format(buyer_address))

        buyer_state = models.get_participant_state(buyer_address)

        ticket_address = exchange_tx.sender.assets[0]
        buyer_state.add_unused_ticket(ticket_address)
        updated_buyer_state = buyer_state.to_state()

        logger.debug("buyer state update is prepared.")

        return protos.ResponseUpdateState(
            code=0,
            assets=[updated_event_state],
            states=[updated_buyer_state],
        )
    else:
        logger.error(
            "Received an update_reqeust for ExchangeTx not recognized.",
        )

    return protos.ResponseUpdateState(code=0)


exchange_tx_handler = forge_helper.TxHandler(
    'fg:t:exchange',
    exchange_verify,
    exchange_update,
)

sdk.init_server([exchange_tx_handler])



from forge import protos
from forge.utils import utils


class RpcState:
    def __init__(self, chan):
        self.stub = protos.StateRpcStub(chan)

    def get_account_state(self, req):
        """
        RPC call to get account state.

        Parameters
        ----------
        req: stream RequestGetAccountState

        Returns
        -------
        stream ResponseGetAccountState

        """

        def to_req(item):
            if isinstance(item, protos.RequestGetAccountState):
                return item
            else:
                kwargs = {
                    'address': item.get('address'),
                    'keys': item.get('keys', []),
                    # 'height': item.get('height', ''),
                }
                return protos.RequestGetAccountState(**kwargs)

        requests = utils.to_iter(to_req, req)

        return self.stub.get_account_state(requests)

    def get_asset_state(self, req=None):
        """
        RPC call to get asset state.

        Parameters
        ----------
        req: stream RequestGetAssetState

        Returns
        -------
        stream ResponseGetAssetState

        """

        def to_req(item):
            if isinstance(item, protos.RequestGetAssetState):
                return item
            else:
                kwargs = {
                    'address': item.get('address'),
                    'keys': item.get('keys', []),
                    # 'app_hash': item.get('app_hash', ''),
                }
                req = protos.RequestGetAssetState(**kwargs)
                return req

        requests = utils.to_iter(to_req, req)

        return self.stub.get_asset_state(requests)

    def get_stake_state(self, req=None):
        """
        RPC call to get stake state.

        Parameters
        ----------
        req: stream RequestGetStakeState

        Returns
        -------
        stream ResponseGetChannelState

        """

        def to_req(item):
            if isinstance(item, protos.RequestGetStakeState):
                return item
            else:
                kwargs = {
                    'address': item.get('address'),
                    'keys': item.get('keys', []),
                    # 'app_hash': item.get('app_hash', ''),
                }
                return protos.RequestGetStakeState(**kwargs)

        requests = utils.to_iter(to_req, req)
        return self.stub.get_stake_state(requests)

    def get_forge_state(self, keys='', app_hash='', req=None):
        """
        RPC call to get forge state.

        Parameters
        ----------
        req: RequestGetForgeState

        Returns
        -------
        ResponseGetForgeState

        """
        if req is not None:
            return self.stub.get_forge_state(req)
        else:
            req_kwargs = {
                'keys': keys,
                # 'app_hash': app_hash,
            }
            return self.stub.get_forge_state(
                protos.RequestGetForgeState(**req_kwargs),
            )
